# Kapitel 10

from math import *

# Selbst gebaute komplexe Zahlen

def makeComplex(a,b): return [a,b]
def re(z): return z[0]
def im(z): return z[1]
def betrag(z): return sqrt(im(z)**2+re(z)**2)
def add(z1,z2): return makeComplex(re(z1)+re(z2),im(z1)+im(z2))
def sub(z1,z2): return makeComplex(re(z1)-re(z2),im(z1)-im(z2))
def mul(z1,z2): return makeComplex(re(z1)*re(z2)-im(z1)*im(z2),
                                   re(z1)*im(z2)+im(z1)*re(z2))
def div(z1,z2): return mul(z1,kehrwert(z2))
def kehrwert(z):
    q=re(z)**2+im(z)**2
    return makeComplex(re(z)/float(q),-im(z)/float(q))
def asString(z): return str(re(z))+"+"+str(im(z))+"j"

z1=makeComplex(4,8)
z2=makeComplex(3,-2)
Z1=4+8j
Z2=3-2j
print("Summe="+asString(add(z1,z2))+" Kontrolle: "+str(Z1+Z2))
print("Quotient="+asString(div(z1,z2))+" Kontrolle: "+str(Z1/Z2))


# Vektoren
def vbetrag(v):
    s=0
    for x in v: s+=x**2
    return sqrt(s)
# eleganter ist:
def vbetrag(v): return sqrt(sum([x**2 for x in v]))

def vadd(a,b):
    if len(a)!=len(b): raise Exception("Vektoren ungleicher Laenge!")
    return [ x+y for (x,y) in zip(a,b) ]
def vsub(a,b):
    if len(a)!=len(b): raise Exception("Vektoren ungleicher Laenge!")
    return [ x-y for (x,y) in zip(a,b) ]
def vdot(a,b): # Skalarprodukt
    if len(a)!=len(b): raise Exception("Vektoren ungleicher Laenge!")
    return sum([ x*y for (x,y) in zip(a,b) ])
def smul(s,a): # Skalarmultiplikation
    return [ s*x for x in a ]

# Matrizen  [[A11,A12...],[A21,...]]  -----------------------
def mdim(M): return [len(M),len(M[0])]
def makeMatrix(n,m):
    ergebnis=[]
    for i in range(n): ergebnis.append([0]*m)
    return ergebnis
def madd(a,b):
    if len(a)!=len(b): raise Exception("Matrizen ungleicher Laenge!")
    return [ vadd(x,y) for (x,y) in zip(a,b) ]
def msub(a,b):
    if len(a)!=len(b): raise Exception("Matrizen ungleicher Laenge!")
    return [ vsub(x,y) for (x,y) in zip(a,b) ]
def msmul(s,a): # Skalarmultiplikation
    return [ smul(s,x) for x in a ]
def transpose(a):
    [n,m]=mdim(a)
    ergebnis=makeMatrix(m,n)
    for i in range(n):
        for j in range(m):
            ergebnis[j][i]=a[i][j]
    return ergebnis

def mdet(a): # Determinante
    [m,n]=mdim(a)
    if n!=m: raise Exception("Det nur von Quadratmatrizen")
    if n==1: return a[0][0]
    d=0
    for i in range(m):
        d+= (-1)**i * a[i][0]*mdet(submat(a,i,0))
    return d
def submat(a,i,j): # Matrix mit Zeile  i, Spalte j entfernt
    ergebnis=[]
    [n,m]=mdim(a)
    for ii in range(n):
        if ii!=i: ergebnis.append(a[ii][:j]+a[ii][j+1:])
    return ergebnis

def mmul(A,B): # Matrixmultplikation
    [m,n]=mdim(A)
    [n1,k]=mdim(B)
    if n!=n1: raise Exception("Falsche Dimensionen")
    ergebnis=makeMatrix(m,k)
    for i in range(m):
        for j in range(k):
            ergebnis[i][j]=0
            for u in range(n): ergebnis[i][j]+=A[i][u]*B[u][j]
    return ergebnis

def printMat(a):
    for z in a: printZeile(z)
def printZeile(z):
    s=""
    for t in [str(x)+"\t" for x in z]: s=s+t
    print("["+s[0:-1]+"]")

M0=[[2,1,5],[9,0,1],[0,1,-2]]
M1=[[3,2,1,5],[6,9,0,1],[1,0,1,-2]]
M2=[[3,2,1],[6,9,0],[1,0,1]]
printMat(M1)
print("Die transponierte davon")
Mt=transpose(M1)
printMat(Mt)
print("det(M2)=",mdet(M2))
print("Matrix-Multiplkations,Determinaten-Test: ", mdet(M0)*mdet(M2),"=",mdet(mmul(M0,M2)))


def swapRows(A,i,j): return A[:i]+[A[j]]+A[i+1:j]+[A[i]]+A[j+1:]

def rowReduce(A): # Stelle die Zeilennormalform her
    [n,m]=mdim(A)
    for i in range(n):
        if A[i][i]==0: # Pivot-Element Null? Dann Zeilen vertauschen
            for j in range(i+1,n):
                if A[j][k]!=0:
                    A=SwapRows(A,i,j)
                    break
        if A[i][i]==0: # Immer noch null - dann fertig
            return A
        A[i]=smul(1/A[i][i],A[i])
        for j in range(i+1,n):
            A[j]=vsub(A[j],smul(A[j][i],A[i]))
    return A
def istNullVektor(a):
    for x in a:
        if x!=0: return False
    return True
def matSolve(A,b): # Rueckgabe [] bedeutet nicht loesbar
    n=len(A[0])
    if len(A)!=len(b): raise Exception("Dimensionsfehler")
    AE=[z.append(bx) for (z,bx) in zip(A,b)]
    AE=rowReduce(A)
    x=["Unbestimmt"]*n # Das wird die Loesung
    for i in range(len(A)-1,-1,-1):
        zeile=AE[i]
        if istNullVektor(zeile): continue
        if istNullVektor(zeile[:-1]): return []
        if AE[i][i]==0: x[i]="Unbestimmt"
        else:
            r=zeile[-1]
            for j in range(i+1,len(zeile)-1):
                if x[j]!="Unbestimmt": r=r-zeile[j]*x[j]
        x[i]=r/AE[i][i]
    return x

    

                       

