# Kapitel 8 Nullstellen, Ableitungen etc

from math import *

def f(x):
    return x**2 + 3*x - 5

# Bisektion
def bisekt(f,a,b):
    if f(a)*f(b)>0:
        print("Vorzeichen an beiden Stellen gleich - keine Lösung")
        return
    while b-a>0.00001:
        m=(a+b)/2
        if f(m)*f(a)<=0: # Vorzeichenwechsel zwischen a und m
            b=m          # Neues Intervall a..m
        if f(m)*f(b)<=0: # Vorzeichenwechsel zwischen b und m
            a=m
    return (a+b)/2

print("Bisektionsnullstellen von f : x=",bisekt(f,0,10))
print("pi-Bisektion: pi=",bisekt(sin,3,4))

# Ableitungen
def ableitung(f,x):
    h=0.00001
    return (f(x+h)-f(x))/h

# Newton
# f(x+h)=f(x)+h*f'(x)=0 h=-f(x)/f'(x)
def newton(f,x0):
    while abs(f(x0))>1e-15:
        x0+=-f(x0)/ableitung(f,x0)
    return x0

print("Newton: Nullstellen von f : x=",newton(f,0))
print("pi-Newton: pi=",newton(sin,3))

def D(f):
    def ableitung(x):
        h=0.00001
        return (f(x+h)-f(x-h))/(2*h)
    return ableitung

cos2=D(sin)

for x in range(1,5): print(x,"cos(x)=",cos(x),"D(sin)(x)=",cos2(x))


# Automatische Nullstellensuche

def nullAuto(f,a,b):
    n=1000 # Zahl der Probestellen
    nullst=[]
    xalt=a
    for i in range(1,n):
        x=a+(b-a)/n*i
        if f(x)*f(xalt)<=0: # Vorzeichenwechsel gefunden
            nullst.append(bisekt(f,xalt,x))
        if abs(f(a))<0.1: # auffällig kleiner Wert:
            nullst.append(newton(f,x))
        xalt=x
    return nullst

print("Nullstellen von f: ",nullAuto(f,-100,100))
