# Kapitel 9 Optimierung

from math import *

def f(x):
    return x**2 + 3*x + 5


# Minimierung durch Durchlaufen
x0=-5
x1=5
step=0.0001
xmin=x0
for i in range(0,int((x1-x0)/step)):
    x=x0+i*step
    if f(x)<f(xmin):
        xmin=x
print("Minimum bei x=",xmin," f an dieser Stelle ist=",f(xmin))

def mini(f,x0,x1,step):
    xmin=x0
    for i in range(0,int((x1-x0)/step)):
        x=x0+i*step
        if f(x)<f(xmin):
            xmin=x
    return xmin

xmin=mini(cos,0,5,0.0001)
print("Minimum des cos bei x=",xmin," cos an dieser Stelle ist=",cos(xmin))


# Minimierung mit Parabelpassung

def abl1(f,x):
    h=0.000001
    return (f(x+h)-f(x))/h
def abl2(f,x):
    h=0.000001
    return (abl1(f,x+h)-abl1(f,x))/h

def minipar(f,x0):
    while True:
        a=0.5*abl2(f,x0)
        b=abl1(f,x0)-x0*abl2(f,x0)
        c=0.5*(x0**2*abl2(f,x0)-2*x0*abl1(f,x0)+2*f(x0))
        x1=-b/(2*a)
        if abs(x1-x0)<0.0000001: return x1
        x0=x1

print("minipar: min cos, startwert 2 : ", minipar(cos,2))
print("minipar: min cos, startwert 1 : ", minipar(cos,1))

def miniparstabil(f,x0):
    maxiter=100 # Maximal 100 Iterationen
    while maxiter>0:
        maxiter=maxiter-1
        a=0.5*abl2(f,x0)
        if a<0.0001: # keine nach oben geoeffnete Parabel
            abl=abl1(f,x0)
            if abl>0: x1=x0-0.01 # schritt nach links
            else:
                if abl<0: x1=x0+0.01
                else: return x0
        else: # Normaler Iterationsschritt
            b=abl1(f,x0)-x0*abl2(f,x0)
            c=0.5*(x0**2*abl2(f,x0)-2*x0*abl1(f,x0)+2*f(x0))
            x1=-b/(2*a) # Das waere der neue Wert
            x1=x0+(x1-x0)*min(a,1) # Bei kleinem a nicht so weit
        if abs(x1-x0)<0.00000001: return x1
        x0=x1
    # print "miniparstabil Warnung: Maximale Iterationszahl erreicht"
    return x0

print("miniparstabil: min cos, startwert 2 : ", miniparstabil(cos,2))
print("miniparstabil: min cos, startwert 1 : ", miniparstabil(cos,1))


# Minimierung mit Klammerung
# Klammerung [a,b,c] fuer f bedeutet: f(a)>f(b) und f(b)<f(c)
def istKlammer(f,K): return f(K[0])>=f(K[1]) and f(K[1])<=f(K[2]) 
def verbessere(f,K):
    [a,b,c]=K
    if not(istKlammer(f,K)): print(("!!!!!!!!!!!!, keine Klammer:",K))
    if istKlammer(f,[(a+b)/2,b,c]): [a,b,c]=[(a+b)/2,b,c] 
    if istKlammer(f,[a,b,(c+b)/2]): [a,b,c]=[a,b,(c+b)/2] 
    if istKlammer(f,[a,(a+b)/2,b]): [a,b,c]=[a,(a+b)/2,b]
    if istKlammer(f,[b,(b+c)/2,c]): [a,b,c]=[b,(b+c)/2,c]
    return [a,b,c]
def miniKlammer(f,K):
    while K[2]-K[0]>0.00000001:
        K1=verbessere(f,K)
        if K==K1 : return K
        K=K1
    return K
print("miniKlammer:  startwert 1,2,5 : ", miniKlammer(cos,[1,2,5]))

def findeKlammer(f,x0):
    delta=0.01
    f0=f(x0)
    if f(x0+delta)<f0:
        a=x0
        b=x0+delta
        fb=f(b)
        for i in range(1,30):
            if f(b+2**i*delta)>fb: return [a,b,b+2**i*delta]
    if f(x0+delta)>f0:
        b=x0
        c=x0+delta
        fb=f(b)
        for i in range(1,30):
            if f(b-2**i*delta)>fb: return [b-2**i*delta,b,c]
    print("Keine Klammerung gefunden")
    return []
def miniK(f,x0):
    K=findeKlammer(f,x0)
    if K==[]: return []
    else: return miniKlammer(f,K)[1]
print("miniK cos :  startwert 1 : ", miniK(cos,2))



# Mehrdimensional

def f2test(A):
    return (A[0]-5)**2+(A[1]-7)**2

def miniparN(f,x0): # x0 ist Liste der Startwerte
    n=len(x0)
    for k in range(50): # 50 Mal ...
        for i in range(n): # ... in jede Dimension eindim minimieren
            def fi(x):
                x0[i]=x
                return f(x0)
            x0[i]=miniparstabil(fi,x0[i])
    return x0

print("Minimum von f2test mit miniparN : ", miniparN(f2test,[1,1]))

# Mehrdimensional in Gradientenrichtung

def grad(f,x0):
    g=[0]*len(x0) # Das wird der Gradient
    h=0.000001
    f0=f(x0)
    for i in range(len(x0)):
        x0[i]+=h # i-te Variable +h
        g[i]=(f(x0)-f0)/h
        x0[i]+=-h  # i-te Variable zuruecksetzen
    return g

def vbetrag(v): return sqrt(sum([x**2 for x in v]))

def muladd(a,s,b):  # a,b, sind Vektoren (d.h. Listen), s ist Zahl
    # berechnet a+s*b
    ergebnis=[0]*len(a)
    for i in range(len(a)):
        ergebnis[i]=a[i]+s*b[i]
    return ergebnis

def miniparNg(f,x0): # x0 ist Liste der Startwerte
    for k in range(50): # 50 Mal ...
        g=grad(f,x0)
        if vbetrag(g)<0.000001: return x0
        if vbetrag(g)>10:
            for i in range(len(g)): g[i]=10*g[i]/vbetrag(g)
        def fr(s): # Funktion in Richtung g
            return f(muladd(x0,s,g))
        s_optimal=miniK(fr,0) # Hier koennte auch miniparstabil stehen
        if abs(s_optimal)<0.0001: return x0 # Keine nennenswerte Aenderung
        x0=muladd(x0,s_optimal,g)
        # print "   x0=",x0,"g=",g
    return x0

print("Minimum von f2test mit miniparNg: ", miniparNg(f2test,[1,1]))

# Beispiel Fermatpunkt

def fFermat(P):
    [x,y]=P
    return sqrt((x-1)**2+(y-3)**2)+sqrt((x-5)**2+(y-2)**2)+sqrt((x-4)**2+(y-7)**2)
print()
print("Fermatpunkt mit miniparN: ", miniparN(fFermat,[1,1]))
print("Fermatpunkt mit miniparNg: ", miniparNg(fFermat,[1,1]))


# Minimierung mit Nebenbedingung

def miniNeben(f,g,x0,gewicht=100): # f wird minimiert unter der Nebenbedingung g=0
    def f2(L):
        return f(L)+gewicht*g(L)**2
    return miniparN(f2,x0)
def miniNebeng(f,g,x0,gewicht=100): # f wird minimiert unter der Nebenbedingung g=0
    def f2(L):
        return f(L)+gewicht*g(L)**2
    return miniparNg(f2,x0)
def miniNebengI(f,g,x0,gewicht=1): # f wird minimiert unter der Nebenbedingung g=0
    def f2(L):
        return f(L)+gewicht*g(L)**2
    for i in range(20):
        x0=miniparNg(f2,x0)
        gewicht=gewicht*2
    return x0


# a) Welcher Punkt auf der Geraden x+y=10 liegt am naechsten am Kreis mit Radius 2?
def fAbstand(P): return (P[0]**2+P[1]**2-2**2)**2 
def gKurve(P): return P[0]+P[1]-10
print()
print("Opt Kurvenpunkt : ", miniNeben(fAbstand,gKurve,[5,4],100))
print("Opt g Kurvenpunkt : ", miniNebeng(fAbstand,gKurve,[5,4],100))
print("Opt gI Kurvenpunkt : ", miniNebengI(fAbstand,gKurve,[5,4],1))


def fDose(L):
    [r,h]=L
    return 2*pi*r*h+2*pi*r**2

def DosenVolumenBedingung(L):
    [r,h]=L
    return (pi*r**2*h-850)*0.01 


print()
print("Dose mit Nebenbedingung : ", end=' ')
print(miniNeben(fDose,DosenVolumenBedingung,[9,9],10))
print("Dose mit Nebenbedingung g : ", end=' ')
print(miniNebeng(fDose,DosenVolumenBedingung,[9,9],10))
print("Dose mit Nebenbedingung gI : ", end=' ')
print(miniNebengI(fDose,DosenVolumenBedingung,[9,9],10))

    
    
# Haenegnde Kette ********************


def Energie(xi,yi): # xi,yi sind Listen der Art xi=[x1,x2,...]
    # Ergebnis ist Gesamtenergie
    E=0.0
    for i in range(len(xi)-1):
        xa=xi[i]
        xb=xi[i+1]
        ya=yi[i]
        yb=yi[i+1]
        y=(ya+yb)/2 # Mittlere Hoehe
        laenge=sqrt((xa-xb)**2+(ya-yb)**2)
        E+=y*laenge
    return E

def Laenge(xi,yi): # xi,yi sind Listen der Art xi=[x1,x2,...]
    # Ergebnis ist Gesamtlaenge
    L=0.0
    for i in range(len(xi)-1):
        xa=xi[i]
        xb=xi[i+1]
        ya=yi[i]
        yb=yi[i+1]
        L+=sqrt((xa-xb)**2+(ya-yb)**2)
    return L

        
def Kette(P,Q,L,n): # P,Q sind Startkoordinaten, L die Laenge der Kette
    xi=[P[0]]
    yi=[P[1]]
    for i in range(n):
        xi.append(P[0]+(Q[0]-P[0])/n*(i+0.5))
        yi.append(P[1]+(Q[1]-P[1])/n*(i+0.5))
    xi.append(Q[0])
    yi.append(Q[1])
    def f(Yinnen):
        yneu=[yi[0]]+Yinnen+[yi[-1]]
        return Energie(xi,yneu)+10*(Laenge(xi,yneu)-L)**2
    for i in range(1): # Mehr als 1-mal erhoeht die Genauigkeit
        yi=[yi[0]]+miniparNg(f,yi[1:-1])+[yi[-1]]
    return list(zip(xi,yi))
     

from SWGui import *


initSWGui()
n=400
Zeichenfeld=SWCanvas(width=n,height=n)
Zeichenfeld.showCoSys()

N=10
L=15
PA=Zeichenfeld.createCircleW(-8,2,0.4)
PA.setFillColor("red")
PB=Zeichenfeld.createCircleW(8,-3,0.4)
PB.setFillColor("red")
Ps=[Zeichenfeld.createCircleW(0.2*i,0,0.2) for i in range(N)]


dragObj="" # Welches Objekt wird gezogen

def update(event):
    if dragObj=="A": PA.setCenter(event.x,event.y)
    if dragObj=="B": PB.setCenter(event.x,event.y)
    Punkte=Kette(PA.getCenterW(),PB.getCenterW(),L,N)
    for i in range(N): Ps[i].setCenterW(Punkte[i+1])

Zeichenfeld.setDragCommand(update)

def setDragA(event):
    global dragObj
    dragObj="A"
def setDragB(event):
    global dragObj
    dragObj="B"
PA.setClickCommand(setDragA)
PB.setClickCommand(setDragB)

Punkte=Kette(PA.getCenterW(),PB.getCenterW(),L,N)
for i in range(N): Ps[i].setCenterW(Punkte[i+1])





    
