# Kapitel 9b Optimierung

from math import *


# Das Folgende sind die Minimierungsfunktionen aus Kapitel 9
def abl1(f,x):
    h=0.000001
    return (f(x+h)-f(x))/h
def abl2(f,x):
    h=0.000001
    return (abl1(f,x+h)-abl1(f,x))/h

def minipar(f,x0):
    while True:
        a=0.5*abl2(f,x0)
        b=abl1(f,x0)-x0*abl2(f,x0)
        c=0.5*(x0**2*abl2(f,x0)-2*x0*abl1(f,x0)+2*f(x0))
        x1=-b/(2.0*a)
        if abs(x1-x0)<0.0000001: return x1
        x0=x1

def miniparstabil(f,x0):
    maxiter=100 # Maximal 100 Iterationen
    while maxiter>0:
        maxiter=maxiter-1
        a=0.5*abl2(f,x0)
        if a<0.0001: # keine nach oben geoeffnete Parabel
            abl=abl1(f,x0)
            if abl>0: x1=x0-0.01 # schritt nach links
            else:
                if abl<0: x1=x0+0.01
                else: return x0
        else: # Normaler Iterationsschritt
            b=abl1(f,x0)-x0*abl2(f,x0)
            c=0.5*(x0**2*abl2(f,x0)-2*x0*abl1(f,x0)+2*f(x0))
            x1=-b/(2.0*a) # Das waere der neue Wert
            x1=x0+(x1-x0)*min(a,1) # Bei kleinem a nicht so weit
        if abs(x1-x0)<0.00000001: return x1
        x0=x1
    # print "miniparstabil Warnung: Maximale Iterationszahl erreicht"
    return x0

def grad(f,x0):
    g=[0]*len(x0) # Das wird der Gradient
    h=0.000001
    f0=f(x0)
    for i in range(len(x0)):
        x0[i]+=h # i-te Variable +h
        g[i]=(f(x0)-f0)/h
        x0[i]+=-h  # i-te Variable zuruecksetzen
    return g

def vbetrag(v): return sqrt(sum([x**2 for x in v]))

def muladd(a,s,b):  # a,b, sind Vektoren (d.h. Listen), s ist Zahl
    # berechnet a+s*b
    ergebnis=[0]*len(a)
    for i in range(len(a)):
        ergebnis[i]=a[i]+s*b[i]
    return ergebnis

def miniparNg(f,x0): # x0 ist Liste der Startwerte
    for k in range(50): # 50 Mal ...
        g=grad(f,x0)
        if vbetrag(g)<0.000001: return x0
        if vbetrag(g)>10:
            for i in range(len(g)): g[i]=10*g[i]/vbetrag(g)
        def fr(s): # Funktion in Richtung g
            return f(muladd(x0,s,g))
        s_optimal=miniparstabil(fr,0) # Hier koennte auch miniparstabil stehen
        if abs(s_optimal)<0.0001: return x0 # Keine nennenswerte Aenderung
        x0=muladd(x0,s_optimal,g)
        # print "   x0=",x0,"g=",g
    return x0

# ***************************************************************


# Schnellste Rutschbahn: Brachistochrone ********************


def Zeit(xi,yi): # xi,yi sind Listen der Art xi=[x1,x2,...]
    # Ergebnis ist Gesamtzeit, wenn Anfangs Geschwindigkleit =0
    t=0.0
    g=9.81  # Erdbeschleunigung
    for i in range(len(xi)-1):
        xa=xi[i]
        xb=xi[i+1]
        ya=yi[i]
        yb=yi[i+1]
        y=(ya+yb)/2.0 # Mittlere Hoehe
        v=sqrt(2*9.81*(yi[0]-y)) # aus 1/2*m*v*v=m*g*h
        strecke=sqrt((xa-xb)**2+(ya-yb)**2)
        t+=strecke/v
    return t

        
def Brachsito(P,Q,n): # P,Q sind Startkoordinaten
    # Anfangsweg: Lineare Rutschbahn:
    xi=[P[0]]
    yi=[P[1]]
    for i in range(n):
        xi.append(P[0]+(Q[0]-P[0])/float(n)*(i+0.5))
        yi.append(P[1]+(Q[1]-P[1])/float(n)*(i+0.5))
    xi.append(Q[0])
    yi.append(Q[1])
    def f(Yinnen): return Zeit(xi,[yi[0]]+Yinnen+[yi[-1]])
    for i in range(1): # Mehr als 1-mal erhoeht die Genauigkeit
        yi=[yi[0]]+miniparNg(f,yi[1:-1])+[yi[-1]]
    return list(zip(xi,yi))

# print Brachsito([-8,8],[8,-3],10)
                

from SWGui import *


initSWGui()
n=400
Zeichenfeld=SWCanvas(width=n,height=n)
Zeichenfeld.showCoSys()

N=10
PA=Zeichenfeld.createCircleW(-8,2,0.4)
PA.setFillColor("red")
PB=Zeichenfeld.createCircleW(8,-3,0.4)
PB.setFillColor("red")
Ps=[Zeichenfeld.createCircleW(0.2*i,0,0.2) for i in range(N)]


dragObj="" # Welches Objekt wird gezogen

def update(event):
    if dragObj=="A": PA.setCenter(event.x,event.y)
    if dragObj=="B": PB.setCenter(event.x,event.y)
    Punkte=Brachsito(PA.getCenterW(),PB.getCenterW(),N)
    for i in range(N): Ps[i].setCenterW(Punkte[i+1])

Zeichenfeld.setDragCommand(update)

def setDragA(event):
    global dragObj
    dragObj="A"
def setDragB(event):
    global dragObj
    dragObj="B"
PA.setClickCommand(setDragA)
PB.setClickCommand(setDragB)

Punkte=Brachsito(PA.getCenterW(),PB.getCenterW(),N)
for i in range(N): Ps[i].setCenterW(Punkte[i+1])





    
