# -*- coding: cp1252 -*-
# SWGui.py - ein einfaches GUI-Modul (Version 6)

# AENDERUNGEN gegenueber Version 5
# setCoords von SWCLine in SWCElement uebertragen
# startSWGui entfernt
# SWText erhaelt bei Aufruf mit undo=1 einen undo und einen redo Button
#   Vielleicht muesste man da nochmal drauf schauen wg. Layout (evtl.
#   uebergeortdeneter Frame?)

# -------------------------------------------

# Aenderungen gegenueber Version 4
# SWCElement setzt mycanvas schon in __init__ (spart aber keine Zeile)
# SWCElement hat move
# SWCElement hat stackUp, stackDown
# Bug in Rueckgabe von Checkbutton und Buttonlist beseitigt
# SWCLine: Koordinaten koennen als Liste oder einzelen angegeben werden
# Bei Farben koennen jetzt auch Listen angegeben werden, zB [255,0,0]

from tkinter import *
import _thread
import string
import math
import time
import turtle
import tkinter.filedialog
import tkinter.colorchooser


try:
    import Image
    import ImageTk
    thisSystemHasPIL=True
except Exception :
    thisSystemHasPIL=False
    

tkroot=None
tkthread=0

def initSWGui(w=600,h=400):
    """ Initialisiert SWGui

    SWGUi luft danach in einem eigenen Thread."""
    global tkroot,tkthread
    tkroot=Tk()
    tkthread=_thread.start_new_thread(startSWGuiProc,())
    
def startSWGuiProc():
    global tkroot
    tkroot.update()
    tkroot.mainloop()

def quitSWGui():
    global tkthread,tkroot
    tkroot.quit()


def sleep(x): return time.sleep(x)

def isNumericList(L):
    if not(isinstance([1,2,3],list)): return False
    for x in L:
        if not(isinstance(x,int) or isinstance(x,float)): return False
    return True

class SWFrameHori(Frame):
    """Erzeugt einen Rahmen, der horizontal angeordnete Widgets aufnimmt"""
    def __init__(self,container=tkroot):
        Frame.__init__(self,container)
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()

class SWFrameVert(Frame): 
    """Erzeugt einen Rahmen, der horizontal angeordnete Widgets aufnimmt"""
    def __init__(self,container=tkroot):
        Frame.__init__(self,container)
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()

class SWButton(Button):
    def __init__(self, container=tkroot,text='', command=None, **options):
        Button.__init__(self,container,text=text,command=command)
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()   
    def setText(self,s):
        """Setzt den Text eines Buttons

        s kann ein beliebiger String sein"""
        self.configure(text=s)
    def setCommand(self,s):
        self.configure(command=s) 

class SWCheckbutton(Checkbutton): # select(), deselect() von  Button 
    def __init__(self, container=tkroot,text='', command=None, **options):
        self.var=IntVar()
        Checkbutton.__init__(self,container,text=text,command=command,variable=self.var)
        self.deselect()
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()
    def setText(self,s):
        self.configure(text=s) 
    def setCommand(self,s):
        self.configure(command=s)
    def getState(self): return self.var.get()==1

class SWRadiobuttonList():  
    def __init__(self, textList,container=tkroot, command=None):
        self.var=IntVar()
        self.buttons=[]
        for i in range(len(textList)):
            self.buttons.append(Radiobutton(container,variable=self.var,text=textList[i],value=i))
        if isinstance(container, SWFrameHori):
            for b in self.buttons: b.pack(side=LEFT)
        else:
            for b in self.buttons: b.pack()
    def setText(self,i,s):
        self.buttons[i].configure(text=s) 
    def setCommand(self,s):
        for b in self.buttons: b.configure(command=s)
    def getState(self): return self.var.get()


class SWLabel(Label):
    def __init__(self, container=tkroot,text='', **options):
        Label.__init__(self,container,text=text)
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()
    def setText(self,s):
        self.configure(text=s) 
    def setWidth(self,w):
        self.configure(width=w) 
    def setHeight(self,h):
        self.configure(height=h)
    def delete(self): self.mycanvas.delete(self.mytk)

class SWEntry(Entry): # get() liefert Text
    def __init__(self, container=tkroot,command=None,text='', **options):
        Entry.__init__(self,container,text=text)
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()
        self.setCommand(command)
    def setText(self,s):
        self.delete(0,END)
        self.insert(0,s)
    def getText(self):
        return Entry.get(self)
    def setCommand(self,cmd):
        if cmd!=None: self.bind("<Return>", cmd)

class SWText(Text): # width: Breite in Spalten, height: Hoehe in Zeilen
    def __init__(self,container=tkroot,height=5,width=20,undo=0):
        Text.__init__(self,container,height=5,width=20,undo=undo)
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()
        if undo==1:
            undoFrame = SWFrameHori()
            undoButton = SWButton(container=undoFrame, text="undo", command=self.edit_undo)
            redoButton = SWButton(container=undoFrame, text="redo", command=self.edit_redo)
    def getText(self):
        return Text.get(self,1.0,END)
    def addText(self,text,pos=END):
        Text.insert(self,pos,text)
    def setText(self,text):
        self.clear()
        self.addText(text)
    def clear(self):
        self.delete(1.0,END)

# ---------- Canvas ------------------------------------------------
def canvasClickHandler(self,event):
    if not(self.ClickHandler==None): self.ClickHandler(event)
    closeobjs=list(self.find_closest(event.x,event.y))
    for [obj,cmd] in self.ClickElementHandlers:
        for co in closeobjs:
            if obj.mytk==co:
                d2=(obj.getX()-event.x)**2+(obj.getY()-event.y)**2
                if obj.hitRadius**2>=d2:
                    event.object=obj
                    cmd(event)
def canvasReleaseHandler(self,event):
    if not(self.ReleaseHandler==None): self.ReleaseHandler(event)
    closeobjs=list(self.find_closest(event.x,event.y))
    for [obj,cmd] in self.ReleaseElementHandlers:
        for co in closeobjs:
            if obj.mytk==co:
                d2=(obj.getX()-event.x)**2+(obj.getY()-event.y)**2
                if obj.hitRadius**2>=d2:
                    event.object=obj
                    cmd(event)
def canvasDragHandler(self,event):
    if not(self.DragHandler==None): self.DragHandler(event)
    closeobjs=list(self.find_closest(event.x,event.y))
    for [obj,cmd] in self.DragElementHandlers:
        for co in closeobjs:
            if obj.mytk==co:
                d2=(obj.getX()-event.x)**2+(obj.getY()-event.y)**2
                if obj.hitRadius**2>=d2:
                    event.object=obj
                    cmd(event)
        
class SWCanvas(Canvas):
    def __init__(self,container=tkroot,width=300,height=200):
        Canvas.__init__(self, container,width=width,height=height)
        self.viewport=[-10,10,-10,10] # [xmin,xmax,ymin,ymax]
        self.kosys=[]
        self.turtlescreen=None
        self.ClickHandler=None
        self.ClickElementHandlers=[]
        self.ReleaseHandler=None
        self.showCoSysFlag=False
        self.ReleaseElementHandlers=[]
        self.DragHandler=None
        self.DragElementHandlers=[]
        self.AllObjects=[]
        self.bind("<Button-1>",lambda event: canvasClickHandler(self,event))
        self.bind("<ButtonRelease-1>",lambda event: canvasReleaseHandler(self,event))
        self.bind("<B1-Motion>",lambda event: canvasDragHandler(self,event))
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()
    def setClickCommand(self,cmd): self.ClickHandler=cmd
    def setReleaseCommand(self,cmd): self.ReleaseHandler=cmd
    def setDragCommand(self,cmd): self.DragHandler=cmd
    def addClickCommand(self,obj,cmd):
        self.ClickElementHandlers.append([obj,cmd])
    def addReleaseCommand(self,obj,cmd): self.ReleaseElementHandlers.append([obj,cmd])
    def addDragCommand(self,obj,cmd): self.DragElementHandlers.append([obj,cmd])
    def getTurtle(self): 
        if self.turtlescreen==None: self.turtlescreen=turtle.TurtleScreen(self)
        return turtle.RawTurtle(self.turtlescreen)
    def setViewport(self,L):
        if not(isNumericList(L)): raise Exception("Dimension vector expected")
        self.viewport=L
        for obj in self.AllObjects:
            if obj.setByWorldCoordinates:
                obj.updateAfterDimChange()
        if self.showCoSysFlag:
            self.deleteCoSys()
            self.showCoSys()
    def W2S(self,xy):
        [xw,yw]=xy
        [xmin,xmax,ymin,ymax]=self.viewport
        return [int(float(self.cget("width"))*(xw-xmin)/float(xmax-xmin)),
                int(float(self.cget("height"))*(ymax-yw)/float(ymax-ymin))]
    def S2W(self,xy):
        [xs,ys]=xy
        [xmin,xmax,ymin,ymax]=self.viewport
        return [xmin+(xmax-xmin)*xs/float(self.cget("width")),
                ymax-(ymax-ymin)*ys/float(self.cget("height"))]
    def deleteCoSys(self):
        for e in self.kosys: e.delete()
        self.showCoSysFlag=False
    def showCoSys(self,color=[0,0,0]):
        self.showCoSysFlag=True
        [xmin,xmax,ymin,ymax]=self.viewport
        if xmin<=0 and xmax>=0: xpos=0
        else:
            if xmin>0: xpos=xmin
            else: xpos=xmax
        self.kosys.append(self.createLineW(xpos,ymin,xpos,ymax))
        if ymin<=0 and ymax>=0: ypos=0
        else:
            if ymin>0: ypos=ymin
            else: ypos=ymax
        self.kosys.append(self.createLineW(xmin,ypos,xmax,ypos))
        dx=0.02*(xmax-xmin)
        dy=0.02*(ymax-ymin)
        self.kosys.append(self.createTextW(xpos+dx,ymin+dy,str(ymin)))
        self.kosys.append(self.createTextW(xpos+dx,ymax-dy,str(ymax)))
        self.kosys.append(self.createTextW(xmin+dx,ypos-dy,str(xmin)))
        self.kosys.append(self.createTextW(xmax+dx,ypos-dy,str(xmax)))
        for i in range(int(xmin),int(xmax)):
            self.kosys.append(self.createLineW(i,ypos-dy/2,i,ypos+dy/2))
        for i in range(int(ymin),int(ymax)):
            self.kosys.append(self.createLineW(xpos-dx/2,i,xpos+dx/2,i))
        for e in self.kosys: e.setColor(color)
    # Die Erzeuger mit W am Ende beziehen sich auf Weltkoordinaten
    def createCircleW(self,x,y,r):
        K=SWCCircle(self,x,y,r,True)
        self.AllObjects.append(K)
 #       K.setCenterW(x,y)
  #      K.setRadiusW(r)
        return K
    def createLineW(self,x1,y1,x2,y2):
        L=SWCLine(self,x1,y1,x2,y2,True)
        self.AllObjects.append(L)
   ##     L.setCoordsW(x1,y1,x2,y2)
        return L
    def createTextW(self,x,y,t):
        L=SWCText(self,1,1,"")
        self.AllObjects.append(L)
        L.setCoordsW([x,y])
        L.setText(t)
        return L
    def createRectangleW(self,x1,y1,x2,y2):
        K=SWCRectangleW(self,x1,y1,x2,y2)
        self.AllObjects.append(K)
        return K
    def createPolygonW(self,*punkte):
        K=SWCPolygonW(self,punkte)
        self.AllObjects.append(K)
        return K
    def createArcW(self,x1,y1,x2,y2):
        K=SWCArcW(self,x1,y1,x2,y2)
        self.AllObjects.append(K)
        return K
    def createOvalW(self,x1,y1,x2,y2):
        K=SWCOvalW(self,x1,y1,x2,y2)
        self.AllObjects.append(K)
        return K
    # Ende der Erzeuger in Weltkoordinaten
    def createCircle(self,x,y,r):
        K=SWCCircle(self,x,y,r)
        self.AllObjects.append(K)
        return K
    def createLine(self,x1,y1,x2,y2):
        K=SWCLine(self,x1,y1,x2,y2)
        self.AllObjects.append(K)
        return K
    def createText(self,x,y,t):
        K=SWCText(self,x,y,t)
        self.AllObjects.append(K)
        return K
    def createRectangle(self,x1,y1,x2,y2):
        K=SWCRectangle(self,x1,y1,x2,y2)
        self.AllObjects.append(K)
        return K
    def createPolygon(self,*punkte):
        K=SWCPolygon(self,punkte)
        self.AllObjects.append(K)
        return K
    def createArc(self,x1,y1,x2,y2):
        K=SWCArc(self,x1,y1,x2,y2)
        self.AllObjects.append(K)
        return K
    def createOval(self,x1,y1,x2,y2):
        K=SWCOval(self,x1,y1,x2,y2)    
        self.AllObjects.append(K)
        return K
    def createImage(self,x,y,img):
        swim=SWCImage(self,x,y,img)
        swim.move(swim.getWidth()/2,swim.getHeight()/2)
        self.AllObjects.append(swim)
        return swim
    def getDim(self):
        return [int(self.cget("width")),int(self.cget("height"))]
    def plotFunction(self,f,x1,x2,color=[0,0,0]):
        n=200
        delta=(x2-x1)/float(n)
        [x0,y0]=[x1,f(x1)]
        for i in range(1,n):
            x=x1+delta*i
            y=f(x)
            L=self.createLineW(x0,y0,x,y)
            L.setColor(color)
            [x0,y0]=[x,y]
    def plotPointsW(self,data,color=[0,0,0]): # data ist Liste von Punkten
        PL=[self.createCircleW(P[0],P[1],0.1) for P in data]
        for P in PL: P.setFillColor(color)
        return PL
    def plotPoints(self,data,color=[0,0,0]): # data ist Liste von Punkten
        PL=[self.createCircle(P[0],P[1],3) for P in data]
        for P in PL: P.setFillColor(color)
        return PL
#    def setClickCommand(self,cmd):
#        self.bind("<Button-1>",cmd)
#    def setReleaseCommand(self,cmd):
#        self.bind("<ButtonRelease-1>",cmd)
#    def setDragCommand(self,cmd):
#        self.bind("<B1-Motion>",cmd)

# ---------- Canvas Elemente ----------------------------------------

class SWCElement():
    def __init__(self,mycanvas):
        self.mycanvas=mycanvas
        self.coordsW=[]
        self.setByWorldCoordinates=False
        self.mytk=None
        self.hitRadius=15 # Fuer Eventhandler
    def updateAfterDimChange(self): # xmin,xmax etc haben sich geaendert...
        pass # Muss in Unterklassen richtig definiert werden
    def setBorderColor(self,col):
        self.mycanvas.itemconfigure(self.mytk,outline=mkColor(col))
    def setColor(self,col):
        self.mycanvas.itemconfigure(self.mytk,fill=mkColor(col))
    def setFillColor(self,col):
        self.mycanvas.itemconfigure(self.mytk,fill=mkColor(col))
    def delete(self): self.mycanvas.delete(self.mytk)
    def stackUp(self):
        self.mycanvas.tag_raise(self.mytk)
    def stackDown(self):
        self.mycanvas.tag_lower(self.mytk)
    def setCoords(self,*rest):
        rest=list(rest)
        if len(rest)==1: coords=rest[0]
        else:
            if len(rest)==4: coords=rest
            else:
                raise Exception("SWCElement.setCoords: wrong arguments")
        self.mycanvas.coords(self.mytk,coords[0],coords[1],coords[2],coords[3])
    def getX(self): return self.mycanvas.coords(self.mytk)[0]
    def getY(self): return self.mycanvas.coords(self.mytk)[1]
    def getCoords(self): return self.mycanvas.coords(self.mytk)
    def move(self,*args):
        args=list(args)
        if len(args)==2: [dx,dy]=args
        else:
            if len(args)==1 and isinstance(args[0],list):
                [dx,dy]=args[0]
            else: raise Exception("SWCElement.move: wrong argument types")
        self.mycanvas.move(self.mytk,dx,dy)
    def moveW(self,*args):
        args=list(args)
        if len(args)==2: [dx,dy]=self.mycanvas.W2S(args)
        else:
            if len(args)==1 and isinstance(args[0],list):
                [dx,dy]=self.mycanvas.W2S(args[0])
            else: raise Exception("SWCElement.move: wrong argument types")
        self.mycanvas.move(self.mytk,dx,dy)
    def setClickCommand(self,cmd): 
        self.mycanvas.addClickCommand(self,cmd)
    def setReleaseCommand(self,cmd):
        self.mycanvas.addReleaseCommand(self,cmd)
    def setDragCommand(self,cmd):
        self.mycanvas.addDragCommand(self,cmd)

def mkColor(*args):
    if len(args)==1:
        if isinstance(args[0],str): return args[0]
        if isinstance(args[0],list):
            if len(args[0])==3:
                return str.replace("#%2x%2x%2x"%(args[0][0],args[0][1],args[0][2])," ","0")
    if len(args)==3:
        return string.replace("#%2x%2x%2x"%(args[0],args[1],args[2])," ","0")
    raise Exception("Color misspecified")
    

class SWCCircle(SWCElement):
    def __init__(self,canvas,x,y,r,WC=False): # WC: Weltkkordinaten
        SWCElement.__init__(self,canvas)
        if not(WC):
            self.mytk=canvas.create_oval(x-r,y-r,x+r,y+r)
            self.radius=r
            self.x=x
            self.y=y
            [xmin,xmax,ymin,ymax]=self.mycanvas.viewport
            [rx,ry]=self.mycanvas.S2W([r,r])
            rw=(rx-xmin+ry-ymin)/2.0
            self.coordsW=self.mycanvas.S2W([x,y])+[rw]
        else:
            [xmin,xmax,ymin,ymax]=self.mycanvas.viewport
            self.coordsW=[x,y,r]            
            [self.x,self.y]=self.mycanvas.W2S(self.coordsW[:2])
            self.radius=self.mycanvas.W2S([self.coordsW[2]+xmin,ymax-self.coordsW[2]])[0]
            self.mytk=canvas.create_oval(self.x-self.radius,self.y-self.radius,self.x+self.radius,self.y+self.radius)
            self.setByWorldCoordinates=True        
    def calcWfromS():
        self.coordsW=self.mycanvas.S2W([self.x,self.y])
        self.coordsW[2]=self.mycanvas.S2W([self.radius,self.radius])[0]
    def calcSfromW():
        [self.x,self.y]=self.mycanvas.W2S(self.coordsW)
        self.radius=self.mycanvas.W2S([self.coordsW[2],self.coordsW[2]])[0]
    def updateAfterDimChange(self): # xmin,xmax etc haben sich geaendert...
        [xmin,xmax,ymin,ymax]=self.mycanvas.viewport
        [self.x,self.y]=self.mycanvas.W2S(self.coordsW[0:2])
        self.radius=self.mycanvas.W2S([self.coordsW[2]+xmin,ymax-self.coordsW[2]])[0]
        self.mycanvas.coords(self.mytk,self.x-self.radius,self.y-self.radius,self.x+self.radius,self.y+self.radius)
    def setRadius(self,r):
        self.radius=r
        self.mycanvas.coords(self.mytk,self.x-self.radius,self.y-self.radius,self.x+self.radius,self.y+self.radius)
        self.setByWorldCoordinates=False
    def setRadiusW(self,r):
        [xmin,xmax,ymin,ymax]=self.mycanvas.viewport
        [rx,ry]=self.mycanvas.W2S([r+xmin,ymax-r])
        rpixel=(rx+ry)/2.0
        self.radius=rpixel
        self.coordsW[2]=r
        self.setByWorldCoordinates=True        
        self.mycanvas.coords(self.mytk,self.x-self.radius,self.y-self.radius,self.x+self.radius,self.y+self.radius)
    def setCenter(self,*args):
        if len(args)==2: (x,y)=args
        else: [x,y]=args[0]
        self.x=x
        self.y=y
        self.coordsW[0:2]=self.mycanvas.S2W([x,y])
        self.mycanvas.coords(self.mytk,int(x-self.radius),int(y-self.radius),int(x+self.radius),int(y+self.radius))
        self.setByWorldCoordinates=False
    def setCenterW(self,*args):
        if len(args)==2: (x,y)=args
        else: [x,y]=args[0]
        self.coordsW[0]=x
        self.coordsW[1]=y
        [x,y]=self.mycanvas.W2S([x,y])
        self.x=x
        self.y=y
        self.mycanvas.coords(self.mytk,int(x-self.radius),int(y-self.radius),int(x+self.radius),int(y+self.radius))
        self.setByWorldCoordinates=True
    def getCenterW(self): return self.coordsW[0:-1]
    def getRadiusW(self): return self.coordsW[-1]
    def getX(self): return self.x
    def getY(self): return self.y
    def getCenter(self): return [self.x,self.y]
    def getRadius(self): return self.radius
    def setCoords(self,*rest):
        self.setByWorldCoordinates=False
        rest=list(rest)
        if len(rest)==1: coords=rest[0]
        else:
            if len(rest)==4: coords=rest
            else:
                raise Exception("SWCCircle.setCoords: wrong arguments")
        self.coords=coords
        if coords[2]-coords[0]==coords[3]-coords[1]:
            self.mycanvas.coords(self.mytk,coords[0],coords[1],coords[2],coords[3])
        else: raise Exception("SWCCircle.setCoords: No Circle")
    def move(self,*args):
        args=list(args)
        if len(args)==2: [dx,dy]=args
        else:
            if len(args)==1 and isinstance(args[0],list):
                [dx,dy]=args[0]
            else: raise Exception("SWCElement.move: wrong argument types")
        self.x+=dx
        self.y+=dy
        self.calcWfromS()
        self.mycanvas.move(self.mytk,dx,dy)
        self.setByWorldCoordinates=False
    def moveW(self,*args):
        args=list(args)
        if len(args)==2: [dx,dy]=args
        else:
            if len(args)==1 and isinstance(args[0],list):
                [dx,dy]=args[0]
            else: raise Exception("SWCElement.move: wrong argument types")
        self.coordsW=[self.coordsW[0]+dx,self.coordsW[1]+dy,self.coordsW[2]]
        self.calcSfromW()
        self.mycanvas.move(self.mytk,dx,dy)
        self.setByWorldCoordinates=True

class SWCLine(SWCElement):
    def __init__(self,canvas,x1,y1,x2,y2,WC=False): # WC es sind Weltkoordinaten 
        SWCElement.__init__(self,canvas)
        if WC:
            self.coordsW=[x1,y1,x2,y2]
            self.coords=self.mycanvas.W2S(self.coordsW[:2])+self.mycanvas.W2S(self.coordsW[2:])
            self.mytk=canvas.create_line(*self.coords)
            self.setByWorldCoordinates=True        
        else:
            self.mytk=canvas.create_line(x1,y1,x2,y2)
            self.coords=[x1,y1,x2,y2]
            self.coordsW=self.mycanvas.S2W(self.coords[:2])+self.mycanvas.S2W(self.coords[2:])
    def calcWfromS():
        self.coordsW=self.mycanvas.S2W(self.coords[:2])+self.mycanvas.S2W(self.coords[2:])
    def calcSfromW():
        self.coords=self.mycanvas.W2S(self.coordsW[:2])+self.mycanvas.W2S(self.coordsW[2:])
    def updateAfterDimChange(self): # xmin,xmax etc haben sich geaendert...
        self.coords=self.mycanvas.W2S(self.coordsW[:2])+self.mycanvas.W2S(self.coordsW[2:])
        self.mycanvas.coords(*[self.mytk]+self.coords)
    def setWidth(self,w): self.mycanvas.itemconfigure(self.mytk,width=w)
    def setCoordsW(self,*rest):
        rest=list(rest)
        if len(rest)==1: coords=rest[0]
        else:
            if len(rest)==4: coords=rest
            else:
                raise Exception("SWCCircle.setCoords: wrong arguments")
        self.coordsW=coords
        self.coords=self.mycanvas.W2S([coords[0],coords[1]])+self.mycanvas.W2S([coords[2],coords[3]])
        self.mycanvas.coords(self.mytk,self.coords[0],self.coords[1],self.coords[2],self.coords[3])
        self.setByWorldCoordinates=True

class SWCRectangle(SWCElement):
    def __init__(self,canvas,x1,y1,x2,y2):
        SWCElement.__init__(self,canvas)
        self.mytk=canvas.create_rectangle(x1,y1,x2,y2)
        self.coords=[x1,y1,x2,y2]
    def setWidth(self,w): self.mycanvas.itemconfigure(self.mytk,width=w)

class SWCPolygon(SWCElement):
   def __init__(self,canvas,*punkte):
      SWCElement.__init__(self,canvas)
      self.mytk=canvas.create_polygon(punkte,fill="",outline="black")
      self.coords=punkte
   def setCoords(self,*rest):   # Hier stimmt noch was nicht
      liste=list(rest)
      if len(liste)==1:
          coords=liste[0]
          self.mycanvas.coords(*[self.mytk]+coords)
      else:
        if len(liste)%2==0: coords=liste
        else:
            raise Exception("SWCPolygon: wrong arguments")
        self.coords=coords
        self.mycanvas.coords(*[self.mytk]+coords)
   def setWidth(self,w): self.mycanvas.itemconfigure(self.mytk,width=w)

class SWCArc(SWCElement):
    def __init__(self, canvas,x1,y1,x2,y2, start=45, extent=90):
        SWCElement.__init__(self,canvas)
        self.coords=[x1,y1,x2,y2]
        self.mytk=canvas.create_arc(self.coords[0],self.coords[1],self.coords[2],self.coords[3], start=45, extent=90)
    def setExtent(self,ex): self.mycanvas.itemconfigure(self.mytk,extent=ex)
    def setStart(self, st): self.mycanvas.itemconfigure(self.mytk,start=st)
    def setStyle(self, sty): self.mycanvas.itemconfigure(self.mytk,style=sty)

class SWCOval(SWCElement):
    def __init__(self,canvas,x1,y1,x2,y2):
        SWCElement.__init__(self,canvas)
        self.mytk=canvas.create_oval(x1,y1,x2,y2)
        self.coords=[x1,y1,x2,y2]
        self.width=abs(x2-x1)
        self.height=abs(y2-y1)
        self.center=[(x2-x1)/2,(y2-y1)/2]
    def setCenter(self,x,y):
        self.coords[0]=int(x-self.width/2)
        self.coords[1]=int(y-self.height/2)
        self.coords[2]=int(x+self.width/2)
        self.coords[3]=int(y+self.height/2)
        self.mycanvas.coords(self.mytk,self.coords[0],self.coords[1],self.coords[2],self.coords[3])
    def setFillColor(self,col): self.mycanvas.itemconfigure(self.mytk,fill=col)


class SWCText(SWCElement):
    def __init__(self,canvas,x,y,text):
        SWCElement.__init__(self,canvas)
        self.mytk=canvas.create_text(x,y,text=text)
        self.coords=[x,y]
    def setCoords(self,coords):
        self.coords=coords
        self.mycanvas.coords(self.mytk,coords[0],coords[1])
    def setCoordsW(self,coords):
        self.coordsW=coords
        coords=self.mycanvas.W2S(coords)
        self.coords=coords
        self.mycanvas.coords(self.mytk,coords[0],coords[1])
    def setText(self,text): self.mycanvas.itemconfigure(self.mytk,text=text)

class SWCImage(SWCElement):
    def __init__(self,canvas,x,y,img):
        global thisSystemHasPIL
        SWCElement.__init__(self,canvas)
        if isinstance(img,str): # Zeichenkette, muss Dateiname sein
            if thisSystemHasPIL:
                self.myPILimgOriginal=Image.open(img).convert("RGBA")
                self.myPILimg=Image.open(img).convert("RGBA")
                pimg=ImageTk.PhotoImage(self.myPILimg)
            else:
                self.myPILimgOriginal=None
                self.myPILimg=None
                pimg = PhotoImage(file=img)
        else:
            if isinstance(img,tuple) and isinstance(img[0],int) and isinstance(img[1],int):
                if thisSystemHasPIL: 
                    self.myPILimgOriginal=Image.new("RGBA", img)
                    self.myPILimg=self.myPILimgOriginal
                    pimg=ImageTk.PhotoImage(self.myPILimg)
                else:
                    self.myPILimgOriginal=None
                    self.myPILimg=None
                    pimg = PhotoImage(width=img[0],height=img[1])
        self.myimg=pimg
        self.width=pimg.width()
        self.height=pimg.height()
        self.mytk=canvas.create_image(x,y,image=pimg)
        self.coords=[x,y]
    def flush(self): # schreibt PIL-Bild nach Photoimage
        if thisSystemHasPIL and self.myPILimg:
            self.myimg.paste(self.myPILimg) 
    def setCoords(self,coords):
        self.coords=coords
        self.mycanvas.coords(self.mytk,coords[0],coords[1])
    def getDim(self):
        return [self.myimg.width(),self.myimg.height()]
    def getWidth(self): return self.myimg.width()
    def getHeight(self): return self.myimg.height()
    def getPixelRGBA(self,x,y):
        x=int(x)
        y=int(y)
        if x<0 or y<0 or x>self.width-1 or y>self.height-1: return [0,0,0,0]
        if thisSystemHasPIL and self.myPILimg:
            return list(self.myPILimg.getpixel((x,y)))
        value = self.myimg.get(x, y)
        #print("getPixelRGBA, Photoimage-Case ", value)
        if type(value) == type(""):
            try: value = int(value)
            except ValueError:
                value = tuple(map(int, str.split(value)))
        return value
    def getPixelRGB(self,x,y):
        x=int(x)
        y=int(y)
        if x<0 or y<0 or x>self.width-1 or y>self.height-1: return [0,0,0]
        if thisSystemHasPIL and self.myPILimg:
            return list(self.myPILimg.getpixel((x,y)))[:-1]
        else:
            L=str.split(self.myimg.get(x,y)," ")
            L=[int(x) for x in L]
            return L
    def getPixel(self,x,y):
        x=int(x)
        y=int(y)
        if x<0 or y<0 or x>self.width-1 or y>self.height-1: return 0
        if thisSystemHasPIL and self.myPILimg:
            a=list(self.myPILimg.getpixel((x,y)))
            return int(sum(a[:-1])/3)
        else:
            r,g,b = self.myimg.get(x,y)
            return int((r+g+b)/3)
    def setPixel(self,x,y,w,flushNow=True):
        x=int(x)
        y=int(y)
        if x<0 or y<0 or x>self.width-1 or y>self.height-1: return
        if thisSystemHasPIL and self.myPILimg:
            self.myPILimg.putpixel((x,y),(w,w,w,255))
            if flushNow: self.flush()
            return
        else:
            f="#%02x%02x%02x" % (w,w,w)
            return self.myimg.put((f,),(int(x),int(y)))
    def setPixelRGB(self,x,y,w,flushNow=True):
        x=int(x)
        y=int(y)
        if x<0 or y<0 or x>self.width-1 or y>self.height-1: return
        if thisSystemHasPIL and self.myPILimg:
            self.myPILimg.putpixel((x,y),tuple(w+[255]))
            if flushNow: self.flush()
            return
        else:
            w=[int(u) for u in w]
            if(len(w)==3): [r,g,b]=w
            else: [r,g,b,alpha]=w
            f="#%02x%02x%02x" % (r,g,b)
            #print("f=",f,"rgb=",(r,g,b),"w=",w,"pos=",[x,y])
            return self.myimg.put((f,),(int(x),int(y)))
    def setPixelRGBA(self,x,y,w,flushNow=True):
        x=int(x)
        y=int(y)
        if x<0 or y<0 or x>self.width-1 or y>self.height-1: return
        if thisSystemHasPIL and self.myPILimg:
            self.myPILimg.putpixel((x,y),tuple(w))
            if flushNow: self.flush()
            return
        else:
            [r,g,b]=w[0:-1]
            f="#%02x%02x%02x" % (r,g,b)
            return self.myimg.put((f,),(int(x),int(y)))
    def rotate(self,angle):
        global thisSystemHasPIL
        if not(thisSystemHasPIL): raise Exception("No PIL")
        self.myPILimg=self.myPILimg.rotate(angle)
        self.myimg=ImageTk.PhotoImage(self.myPILimg)
        self.mycanvas.delete(self.mytk)
        self.mytk=self.mycanvas.create_image(self.coords[0],self.coords[1],
                                      image=self.myimg)
        
# ----- Scale ------------
    
class SWScale(Scale):
    def __init__(self,ori,container=tkroot,minimum=0,maximum=100):
        # ori ist "vertical" oder "horizontal"
        Scale.__init__(self, container)
        self.mycontainer=container
        self.mini=minimum
        self.maxi=maximum
        self.configure(from_=minimum,to=maximum)
        self.configure(orient=ori)
        if isinstance(container, SWFrameHori):
            self.pack(side=LEFT)
        else: self.pack()
    def setCommand(self,cmd):
        self.configure(command=cmd)
        

class SWScaleHori(SWScale):
    def __init__(self,container=tkroot,minimum=0,maximum=100):
        SWScale.__init__(self,"horizontal",container,minimum,maximum)

class SWScaleVert(SWScale):
    def __init__(self,container=tkroot,minimum=0,maximum=100):
        SWScale.__init__(self,"horizontal",container,minimum,maximum)


# --- Dialoge --------------

def SWFileNameDialog(): # funktioniert nicht ?????
    return tkinter.filedialog.askopenfilename()

def SWColorDialog(title='Pick Color'): # Funktioniert auch nicht
        ctuple,cstr = tkinter.colorchooser.askcolor(title=title)
        return list(ctuple)

def SWListChoiceDialog(liste,frage=None): # Waehle ein Element aus einer Liste aus
    """Whlt ein Element aus einer Liste aus

    liste ist eine Liste von Zeichenketten. Zurckgegegen wird das gewhlte Element"""
    # im wesentlichen aus Python Cookbook 
    return ListBoxChoice(tkroot, list=liste,message=frage).returnValue()

class ListBoxChoice(object):
    def __init__(self, master=None, title=None, message=None, list=[]):
        self.master = master
        self.value = None
        self.list = list[:]
        
        self.modalPane = Toplevel(self.master)

        self.modalPane.transient(self.master)
        self.modalPane.grab_set()

        self.modalPane.bind("<Return>", self._choose)
        self.modalPane.bind("<Escape>", self._cancel)

        if title:
            self.modalPane.title(title)

        if message:
            Label(self.modalPane, text=message).pack(padx=5, pady=5)

        listFrame = Frame(self.modalPane)
        listFrame.pack(side=TOP, padx=5, pady=5)
        
        scrollBar = Scrollbar(listFrame)
        scrollBar.pack(side=RIGHT, fill=Y)
        self.listBox = Listbox(listFrame, selectmode=SINGLE)
        self.listBox.pack(side=LEFT, fill=Y)
        scrollBar.config(command=self.listBox.yview)
        self.listBox.config(yscrollcommand=scrollBar.set)
        self.list.sort()
        for item in self.list:
            self.listBox.insert(END, item)

        buttonFrame = Frame(self.modalPane)
        buttonFrame.pack(side=BOTTOM)

        chooseButton = Button(buttonFrame, text="Choose", command=self._choose)
        chooseButton.pack()

        cancelButton = Button(buttonFrame, text="Cancel", command=self._cancel)
        cancelButton.pack(side=RIGHT)

    def _choose(self, event=None):
        try:
            firstIndex = self.listBox.curselection()[0]
            self.value = self.list[int(firstIndex)]
        except IndexError:
            self.value = None
        self.modalPane.destroy()

    def _cancel(self, event=None):
        self.modalPane.destroy()
        
    def returnValue(self):
        self.master.wait_window(self.modalPane)
        return self.value


#---- YesNoDialog
def SWYesNoDialog(frage):
    """Fragt nach einer Ja/Nein-Entscheidung

    frage ist eine Zeichenkette, die den Fragetext angibt"""
    return YesNoChoice(tkroot, title="Yes or No?",message=frage).returnValue()

class YesNoChoice(object):
    def __init__(self, master=None, title=None,message=None):
        self.master = master
        self.value = None
        self.modalPane = Toplevel(self.master)
        self.modalPane.transient(self.master)
        self.modalPane.grab_set()
        self.modalPane.bind("<Return>", self._yes)
        self.modalPane.bind("<Escape>", self._cancel)
        if title:
            self.modalPane.title(title)
        if message:
            Label(self.modalPane, text=message).pack(padx=5, pady=5)
        buttonFrame = Frame(self.modalPane)
        buttonFrame.pack(side=BOTTOM)
        yesButton = Button(buttonFrame, text="Yes", command=self._yes)
        yesButton.pack()
        noButton = Button(buttonFrame, text="No", command=self._no)
        noButton.pack(side=RIGHT)
    def _yes(self, event=None):
        self.value=True
        self.modalPane.destroy()
    def _no(self, event=None):
        self.value=False
        self.modalPane.destroy()
    def _cancel(self, event=None):
        self.value=None
        self.modalPane.destroy()
    def returnValue(self):
        self.master.wait_window(self.modalPane)
        return self.value

#---- Message
def SWMessage(text):
    """Zeit einen Nachrichtentext an"""
    return Message(tkroot, title="!",message=text).returnValue()

class Message(object):
    def __init__(self, master=None, title=None,message=None):
        self.master = master
        self.value = None
        self.modalPane = Toplevel(self.master)
        self.modalPane.transient(self.master)
        self.modalPane.grab_set()
        self.modalPane.bind("<Return>", self._yes)
        if title:
            self.modalPane.title(title)
        if message:
            Label(self.modalPane, text=message).pack(padx=5, pady=5)
        buttonFrame = Frame(self.modalPane)
        buttonFrame.pack(side=BOTTOM)
        yesButton = Button(buttonFrame, text="OK", command=self._yes)
        yesButton.pack()
    def _yes(self, event=None):
        self.value=True
        self.modalPane.destroy()
    def _cancel(self, event=None):
        self.value=None
        self.modalPane.destroy()
    def returnValue(self):
        self.master.wait_window(self.modalPane)
        return self.value


#---- TextInput
def SWTextDialog(frage,test=None,transformer=None):
    """Der Benutzer wird aufgefordert, einen Text einzugeben.

    frage: Der Fragetext, ein String
    test: Optionale Funktion die prft, ob die Eingabe OK ist
    transformer: Optionale Funktion, die den eingegeben String bearbeitet, zB in ein Objekt wandelt"""
    return TextChoice(tkroot, title="Enter",
                      message=frage,test=test,transformer=transformer).returnValue()

def SWIntDialog(frage,test=None):
    """Der Benutzer wird aufgefordert, eine ganze Zahl einzugeben.

    frage: Der Fragetext, ein String
    test: Optionale Funktion die prft, ob die Eingabe OK ist"""
    return TextChoice(tkroot, title="Enter",
                      message=frage,test=test,transformer=int).returnValue()

def SWFloatDialog(frage,test=None):
    """Der Benutzer wird aufgefordert, eine Fliepunktzahl einzugeben.

    frage: Der Fragetext, ein String
    test: Optionale Funktion die prft, ob die Eingabe OK ist"""
    return TextChoice(tkroot, title="Enter",
                      message=frage,test=test,transformer=float).returnValue()

class TextChoice(object):
    def __init__(self, master=None, title=None,message=None,test=None,transformer=None):
        self.master = master
        self.transformer=transformer
        self.test=test
        self.value = None
        self.modalPane = Toplevel(self.master)
        self.modalPane.transient(self.master)
        self.modalPane.grab_set()
        self.modalPane.bind("<Return>", self._yes)
        self.modalPane.bind("<Escape>", self._cancel)
        if title:
            self.modalPane.title(title)
        if message:
            Label(self.modalPane, text=message).pack(padx=5, pady=5)
        buttonFrame = Frame(self.modalPane)
        self.en=SWEntry(container=self.modalPane)
        self.en.bind("<Return>", self._yes)
        self.en.bind("<Escape>", self._cancel)
        buttonFrame.pack(side=BOTTOM)
        self.en.pack()
        yesButton = Button(buttonFrame, text="Yes", command=self._yes)
        yesButton.pack(side=RIGHT)
        noButton = Button(buttonFrame, text="No", command=self._no)
        noButton.pack(side=LEFT)
    def _yes(self, event=None):
        self.value=self.en.getText()
        if type(self.transformer)!=type(None):
            try:
                self.value=self.transformer(self.value)
            except:
                SWMessage("Fehlerhafte Eingabe")
                return
        if type(self.test)==type(None):
            self.modalPane.destroy()
            return
        if self.test(self.value):
            self.modalPane.destroy()
            return
        SWMessage("Fehlerhafte Eingabe")
    def _no(self, event=None):
        self.value=None
        self.modalPane.destroy()
    def _cancel(self, event=None):
        self.value=None
        self.modalPane.destroy()
    def returnValue(self):
        self.master.wait_window(self.modalPane)
        return self.value

